@@ -646,7 +646,7 @@ static bt_status_t get_characteristic(int conn_id, btgatt_srvc_id_t *srvc_id,
 	cmd->conn_id = conn_id;
 
 	srvc_id_to_hal(&cmd->srvc_id, srvc_id);
-	len += sizeof(cmd->srvc_id);
+	cmd->number = 0;
 
 	if (start_char_id) {
 		gatt_id_to_hal(&cmd->gatt_id[0], start_char_id);
@@ -676,14 +676,14 @@ static bt_status_t get_descriptor(int conn_id, btgatt_srvc_id_t *srvc_id,
 	cmd->number = 1;
 
 	if (start_descr_id) {
-		gatt_id_to_hal(&cmd->gatt_id[1], char_id);
+		gatt_id_to_hal(&cmd->gatt_id[1], start_descr_id);
 		len += sizeof(cmd->gatt_id[1]);
 		cmd->number++;
 	}
 
 	return hal_ipc_cmd(HAL_SERVICE_ID_GATT,
 					HAL_OP_GATT_CLIENT_GET_DESCRIPTOR,
-					len, &cmd, 0 , NULL, NULL);
+					len, cmd, 0 , NULL, NULL);
 }
 
 static bt_status_t read_characteristic(int conn_id, btgatt_srvc_id_t *srvc_id,
@@ -1065,7 +1065,7 @@ static bt_status_t send_indication(int server_if, int attribute_handle,
 
 	return hal_ipc_cmd(HAL_SERVICE_ID_GATT,
 					HAL_OP_GATT_SERVER_SEND_INDICATION,
-					cmd_len, &cmd, 0, NULL, NULL);
+					cmd_len, cmd, 0, NULL, NULL);
 }
 
 static bt_status_t send_response(int conn_id, int trans_id, int status,
@@ -1084,7 +1084,7 @@ static bt_status_t send_response(int conn_id, int trans_id, int status,
 
 	return hal_ipc_cmd(HAL_SERVICE_ID_GATT,
 					HAL_OP_GATT_SERVER_SEND_RESPONSE,
-					cmd_len, &cmd, 0, NULL, NULL);
+					cmd_len, cmd, 0, NULL, NULL);
 }
 
 static bt_status_t init(const btgatt_callbacks_t *callbacks)
